/*
 * led-power - main.c
 *
 * Copyright (C) 2017, 2018  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <adc.h>
#include <jiffies.h>
#include <watchdog.h>
#include <limits.h>
#include <io.h>
#include <stdio.h>

/*
 * Temperature control set points.
 *
 * led-power uses 200 W MOSFET (SPP80P06P) for the current limiter.
 * The MOSFET is powerful enough. Its case temperature is limited
 * to 125 C due to PCB temperature limit. At 125 C it is limited
 * to 125 W, more than required. It is assumed that the ambient
 * temperature is less than 75 C.
 *
 * The transistor is turned off transistor if its temperature is above
 * T_HIGH over ambient temperature; it is turned on if its temperature
 * drops below T_LOW.
 */
#define T_HIGH		60000
#define T_LOW		10500

/*
 * Maximal DC input voltage for normal operation, 24.0 V +/- 10%,
 * + 5% measurement error, 27.8 V. Plus 7.2% measurement
 * error, again, 29.19 V max.
 */
#define VIN_MAX		27800

/*
 * special short-circuit protection
 */
#define T_SHORT		25000
#define V_SHORT		1000

/*
 * the initial temperature
 */
#define T_INIT		T_HIGH

/*
 * Parameters for worst-case current estimation, maximal current
 * for 75C from SPICE model + 5% margin. V_MAX based on minimal resistance,
 * 5% for resistors, 10% for measurement error.
 */
#define V_D	0	/* mV, voltage drop on other elements */
#define I_MAX	2200	/* mA, saturation current */
#define V_MAX	705	/* mV, minimal voltage for saturation */

#define I_SHIFT	4

/*
 * estimate_ic - estimate the input current
 * @vd: voltage difference, in millivolts (mV).
 *
 * This function returns the estimated current in milliamps (mA).
 *
 * We use simple linear model with two saturations. Below V_D
 * the current is equal to zero, because the voltage is too low
 * for series diode. Above V_MAX the current is constant.
 */
static unsigned int estimate_ic(unsigned int vd)
{
	unsigned int ic = 0;

	if (vd <= V_D)
		return 0;
	if (vd >= V_MAX)
		return I_MAX;
	vd -= V_D;
	ic = (((I_MAX << I_SHIFT) + (V_MAX - V_D) / 2) / (V_MAX - V_D))
		* vd >> I_SHIFT;
	return ic;
}

/*
 * estimate_power - estimate the power on BD244
 * @vd: voltage difference, in millivolts (mV).
 * @ic: collector current, in milliamps (mA).
 *
 * This function returns the estimated power in milliwats (mW).
 */
unsigned int estimate_power(unsigned int vd, unsigned int ic)
{
	unsigned int v_other;
	unsigned char v;
	unsigned char i;

	if (vd > VIN_MAX)
		return 62050;

	/*
	 * voltage on other elements - two parallel 0.75 resistors,
	 * 0.35625 ohm min, approximated with 0.34375 ohm.
	 */
	v_other = (ic >> 2) + (ic >> 4) + (ic >> 5);
	if (v_other >= vd)
		v_other = vd;
	vd -= v_other;

	/* assume at least 128 mV on MOSFET */
	if (vd < 128)
		vd = 128;

#if 0
	return ((unsigned long) vd * ic) >> 11;
#else
	i = (ic + 15) >> 4;
	if (vd < 4080) {
		v = (vd + 15) >> 4;
		return (v * i) >> 3;
	}
	v = (vd + 127) >> 7;
	return (v * i);
#endif
}

#define T_MAX	(65536 - 1024)

static unsigned int q1_thermal_model(unsigned int p)
{
	static unsigned int temp = T_INIT;
	static unsigned int temp_e;
	static unsigned char p_e;

	/*
	 * BD244 temperature control
	 *
	 * For low voltages the BD244 is maximal current-limited,
	 *
	 * BC244 thermal model
	 *
	 *         R P
	 * T = ----------
	 *      RC s + 1
	 *
	 * T - temperature over ambient temperature (75 C assumed),
	 *     in millikelvins (mK)
	 * C - thermal capacitance, 0.54 J/K, from Fairchild AN-7516
	 * 	+ 0.053 (conservative screw + PCB capacitance)
	 * R - thermal resistivity, assumed 21.8 K/W based on
	 *	PCB FDM thermal model + 28% safety margin
	 * P - transistor power
	 *
	 * After discretization using backward differences with HZ=61.035
	 * frequency:
	 *
	 *         R P
	 * T = ----------------
	 *      R C (1 - q) HZ + 1
	 *
	 * where q = z^-1.
	 *
	 * (R C HZ + 1) T - R C HZ q T = R P
	 * (R C HZ + 1) T = R C HZ q T + R P
	 * (R C HZ + 1) T = (R C HZ + 1) q T - q T + R P
	 *
	 *               q T           R P
	 * T = q T - ----------- + ------------
	 *            R C HZ + 1    R C HZ + 1
	 *
	 * R C HZ + 1 = 1022.1, approximated with 1024
	 * R / (R C HZ + 1) = 0.027301, approximated with
	 *	1/2 (1/16 - 1/128) = 0.027344
	 *
	 * To increase precision we use simple error shaper, the previous
	 * quantization error is added to input in the next sample.
	 */
	temp_e += temp;
	temp -= temp_e >> 10;
	temp_e &= 0x3ff;

	p += p_e;
	temp += 6 * (p >> 7);	/* p/2 */
	p_e = p & 0x7f;
//	temp += (p >> 4);	/* p/2 */
//	p_e = p & 0x0f;
	if (temp > T_MAX)
		temp = T_MAX;
	return temp;
}

void wdt_init(void)
{
	watchdog_init();
}

int main(void)
{
	unsigned char t;
	unsigned int vin;
	unsigned int vout;
	unsigned int ic;
	unsigned int p;
	unsigned char enabled = 0;
	unsigned int temp;

	wdt_init();
	io_init();
	adc_init();
	jiffies_init();

	wdt_reset();

	t = get_jiffies();
	for (;;) {
		wdt_reset();

		/*
		 * We implement discrete-time control, the I/O phase
		 * should be as short as possible.
		 */
		vin = vin_get();
		vout = vout_get();
		switch_set(enabled);

		/*
		 * estimate the current and power dissipated on transistor
		 */
		ic = 0;
		p = 0;
		if (enabled) {
			if (vin > vout) {
				unsigned int vd = vin - vout;

				ic = estimate_ic(vd);
				p = estimate_power(vd, ic);
			}

			/* to avoid zero power dissipation */
			p++;
		}

		/*
		 * estimate the transistor temperature
		 */
		temp = q1_thermal_model(p);

		/*
		 * simple controller
		 */
		if (vin > VIN_MAX)
			enabled = 0;
		else if (temp > T_HIGH)
			enabled = 0;
		else if (temp > T_SHORT && vout < V_SHORT)
			enabled = 0;
		else if (temp < T_LOW)
			enabled = 1;

		logger(vin, vout, ic, 2 * p, temp, enabled);
		io_xmit(!enabled);

		t++;
		sleep_to(t);
	}
}
